
* Factor & Extension Analysis Program  (SPSS version).

* primary sources
   Gorsuch, R. L. (1997a). Exploratory factor analysis: Its role 
   in item analysis. Journal of Personality Assessment, 68, 532-560.
   
*  Gorsuch, R. L. (1997b). New procedure for extension analysis in 
   exploratory factor analysis. Ed & Psych Meas, 57, 725-740.

* To specifiy your options, repond to the queries at the 
  start of the program.

* Prepare a raw data matrix for analysis,
  where rows = cases, & columns = variables;
  Cases with missing values are not permitted in the data file;
  The core variables must be the first x variables in the data file;
  The extension variables must appear after (to the right of) 
  the core variables in the data file.

* For the trial analyses, the program was set to read the raw data 
  file named "trial.dat", and then process the active SPSS data file -- 
  see the "GET data" command near the start of the program.  However,
  any SPSS system file can be specified on the GET statement;
  Alternatively, the READ statement (a MATRIX--END MATRIX command)
  can also be used to read a raw data file.

* Make sure to indicate the location of the "trial.dat" file on your Hard Drive.

data list file='Macintosh HD:Latest:ExtensionAnalysis:trial.dat' free  / v1 to v18.


set length=none printback = off  width = 120  mxloops = 9000. 
matrix.

* Enter or read a raw data matrix, where rows = cases, & columns = variables
  Use the following name for the raw data matrix: "data"
  Cases with missing values are not permitted in the data file.

get data / file=* / variables = v1 to v18.


* How many core variables are there? 
  The core variables must be the first x variables in the data file.
compute core = 14.

* How many extension variables are there?
  The extension variables must appear after (to the left of) 
  the core variables in the data file.
compute ext  = 4.

* What kind of extension analysis do you want to run?
  Enter "1" for regular extension analysis
  Enter "2" for higher order factor/extension analysis.
compute kind = 2.

* Specify a procedure for determining the number of factors:
  Enter "1" for # of salient loadings >= abs(.40) (Gorsuch, 1997, J Pers Assess, p 545)
  Enter "2" for parallel analysis of principal components 
            (Zwick & Velicer, 1986, Psych Bull, 99, p 432)
  Enter "3" for parallel analysis of principal axis / common factors
  Enter "4" for for Velicer's MAP test (Velicer, 1976, Psychometrika, 41, 321-327)
  Enter "5" for Standard Error Scree test (Zoski & Jurs, 1996, EPM, p 443)
  Enter "6" for number of eigenvalues > 1
  Enter "7" for a user-selected number of factors (& see the next command).
compute roottest = 1.

* If "7" was specified for roottest, then enter the desired number of factors.
compute nfactors = 4.

* If "7" was specified for roottest, then enter the desired number of factors
  for the higher order analysis (if requested).
compute nfactsho = 2.


* Specify a factor extraction procedure:
  Enter "1" for common factor analysis
  Enter "2" for principal components
  Enter "3" for maximum likelihood factor analysis
  Enter "4" for image analysis.
compute extract = 1.

* Specify a rotation procedure:
  Enter "1" for promax
  Enter "2" for varimax     (not appropriate for higher order analyses)
  Enter "3" for No rotation (not appropriate for higher order analyses).
compute rotate = 1.


* End of required user specifications.


* The following additional specifications can also be made (or ignored).

* Parallel analysis: Enter the # of random data sets.
compute Ndatsets = 10.

* Parallel analysis: Enter the desired percentile eigenvalue (default = 95th).
compute percent = 95.

* Number of Salient Loadings: 
  Enter the minimum abs value for a "salient" loading (default = .40).
compute salvalue = .40.

* Number of Salient Loadings: 
  Enter the minimum # of salient loadings for there to be a "factor" (default = 3).
compute numsals = 3.

* Common Factor analysis: maximum number of iterations (default = 30).
compute itercfa = 3.

* Maximum Likelihood Factor analysis: maximum number of iterations (default = 30).
compute iterml = 30.

* Maximum Likelihood Factor analysis: tolerance for loadings (default = .001).
compute tolerml = .001.

* Promax rotation: the power or k value (default = 3).
compute ppower = 3.

* Promax rotation: 
  Enter "1" for the structure matrix (default, & recommended by Gorsuch, 1997, p 556)
  Enter "2" for the pattern matrix.
compute loadmat = 1.

* End of optional specifications.



compute cases = nrow(data). 
compute nvars = ncol(data).


print  / title='Extension Factor Analysis'.
print cases / title='Number of Cases in the Raw Data Matrix = '.
print nvars / title='Total Number of Variables in the Raw Data Matrix = '.
print core  / title='Number of Core Variables = '.
print ext   / title='Number of Extension Variables = '.


* Improper specification warnings.
do if ( (core+ext) > nvars).
print / title='WARNING: More core and/or extension variables were'. 
print / title='specified than there are variables in the raw data'. 
print / title='matrix: Expect error messages'.
end if.
do if ( nvars > (core+ext) ).
print / title='WARNING: There are more variables in the raw data'. 
print / title='than there are core & extension variables.  The data '. 
print / title='matrix was trimmed to core+extension variables.'.
compute data = data(:,1:(core+ext)).
compute nvars = ncol(data).
end if.


do if   ( roottest = 1).
print / title='Test for Number of Factors:  # Salient Loadings'.
else if ( roottest = 2).
print /title='Test for # of Factors: Parallel Analysis of Prin Components'.
else if ( roottest = 3).
print /title='Test for # of Factors: Parallel Analysis of Common Factors'.
else if ( roottest = 4).
print / title="Test for Number of Factors:  Velicer's MAP test".
else if ( roottest = 5).
print / title='Test for Number of Factors:  Standard Error Scree'.
else if ( roottest = 6).
print / title='Test for Number of Factors:  # Eigenvalues > 1'.
else if ( roottest = 7).
print / title='Test for Number of Factors:  User-Specified'.
end if.

do   if ( extract = 1).
print / title='Factor Extraction Procedure:  Common Factor Analysis'.
else if ( extract = 2).
print / title='Factor Extraction Procedure:  Principal Components'.
else if ( extract = 3).
print / title='Factor Extraction Procedure:  Maximum Likelihood'.
else if ( extract = 4).
print / title='Factor Extraction Procedure:  Image Analysis'.
end if.

do if ( rotate = 1).
print / title='Rotation Procedure:  Promax'.
else if ( rotate = 2).
print / title='Rotation Procedure:  Varimax'.
else if ( rotate = 3).
print / title='Rotation Procedure:  No Rotation'.
end if.

* correlation matrix.
compute zscores = make(nrow(data), ncol(data), -999).
compute n   = nrow(data).
compute nm1 = n - 1.
loop #a = 1 to ncol(data).
compute mean = csum(data(:,#a)) / n.
compute sd = sqrt ( cssq((data(:,#a)- mean)) / nm1  ).
compute zscores(:,#a) = (data(:,#a) - mean) / sd.
end loop.
compute rdata = (1 / (nrow(data)-1) ) * ( t(zscores) * zscores ) .

compute rcore =  rdata(1:core,1:core) .
compute r = rcore.
compute evals1 = eval(r).


* Number of Factors.

* number of salient loadings as the basis for determining the # of factors
  Gorsuch 1997 J Pers Assess p 545.
do if ( roottest = 1).
*  # of eigenvalues > 1.
compute evals = eval(r).
compute nevalgt1 = 0.
loop #a = 1 to nrow(evals).
do if ( evals(#a,1) > 1.0).
compute nevalgt1 = nevalgt1 + 1.
end if.
end loop.
compute nfactors = nevalgt1.
* CFA / PAF  -- from Bernstein p 189   --  smc = from Bernstein p 104.
compute rcfa = r.
compute smc = 1 - (1 &/ diag(inv(rcfa)) ).
loop #a = 1 to itercfa.
call setdiag(rcfa,smc).
call eigen(rcfa,eigvect,eigval).
compute eigval = mdiag(eigval).
compute lding=eigvect(:,1:nfactors)*sqrt(eigval(1:nfactors,1:nfactors)).
compute communal = rssq(lding).
do if ( mmax(abs(communal-smc))<.001 ).
break.
else.
compute smc=communal.
end if.
end loop.
* varimax rotation -- Marcus, 1993, in Reyment & Joreskog, 
  Applied Factor Analysis in the Natural Sciences, CUP.
do if (nfactors > 1).
compute b=lding.
compute n = nrow(lding).
compute nf = ncol(lding).
compute hjsq=diag(lding*t(lding))   .
compute hj=sqrt(hjsq).
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V0=Vtemp.
loop #it=1 to 10 . 
loop #i =1 to nf-1 .
compute jl=#i+1 .
loop #j=jl to nf.
compute xj=lding(:,#i) &/ hj.   
compute yj=lding(:,#j) &/ hj.   
compute uj=xj &* xj-yj &* yj.
compute vj=2*xj &* yj.
compute A=csum(uj).
compute bigB=csum(vj).
compute C=t(uj) * uj - t(vj) * vj.
compute D=2 * t(uj) * vj.
compute num=D-2*A*bigB/n.
compute den=C-(A**2-bigB**2)/n.
compute tan4p=num/den.
compute tan4p = num / den .
compute artanp = ( artan(abs(tan4p)) ) / 4.
do if (num>0 and den>0).
compute phi = artanp.
else if (num<0 and den<0).
compute phi =  -1 * (.7854 - artanp)  .
else if (num<0 and den>0).
compute phi =  -1 * artanp.
else if (num>0 and den<0).
compute phi =  .7854 - artanp .
end if.
compute angle = phi * 180 / 3.1415926  .
do if (  abs(phi)>.00001  ).
compute bigXj=cos(phi)*xj+sin(phi)*yj.
compute bigYj=-sin(phi)*xj+cos(phi)*yj.
compute bj1=bigXj &* hj.
compute bj2=bigYj &* hj.
compute b(:,#i)=bj1.
compute b(:,#j)=bj2.
compute lding(:,#i)=b(:,#i).
compute lding(:,#j)=b(:,#j).
end if.
end loop.
end loop.
compute lding=b.
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V=Vtemp.
do if ( abs(V-V0)<.0001 ).
break.
else.
compute V0=V.
end if.
end loop.
end if.
compute rowmax = rmax(abs(lding)).
compute nfSAL = 0.
loop #lupec = 1 to ncol(lding).
compute nsalynts = 0.
loop #luper = 1 to nrow(lding).
do if ( abs(lding(#luper,#lupec)) >= salvalue and abs(lding(#luper,#lupec)) =
       rowmax(#luper,1) ).
compute nsalynts = nsalynts + 1.
end if.
end loop.
do if ( nsalynts >= numsals ).
compute nfSAL = nfSAL + 1.
end if.
end loop.
compute nfactors = nfSAL.
end if.

* Parallel Analysis program for Principal Components.
do if ( roottest = 2).
compute Ncases   = nrow(data). 
compute Numvars  = nrow(r).
compute evalsr = eval(r).
* computing random data correlation matrices & eigenvalues.
compute evals = make(Numvars,ndatsets,-9999).
compute nm1 = 1 / (ncases-1).
loop #nds = 1 to ndatsets.
compute x = sqrt(2 * (ln(uniform(ncases,Numvars)) * -1) ) &*
            cos(6.283185 * uniform(ncases,Numvars) ).
compute vcv = nm1 * (sscp(x) - ((t(csum(x))*csum(x))/ncases)).
compute d = inv(mdiag(sqrt(diag(vcv)))).
compute evals(:,#nds) = eval(d * vcv * d).
end loop.
* identifying the eigenvalues corresponding to the desired percentile.
compute num = rnd((percent*ndatsets)/100).
compute prcntyle = { t(1:Numvars) }.
loop #root = 1 to Numvars.
compute ranks = rnkorder(evals(#root,:)).
loop #col = 1 to ndatsets.
do if (ranks(1,#col) = num).
compute prcntyle(#root,1) = evals(#root,#col).
break.
end if.
end loop.
end loop.
compute nfPARPCA = 0.
loop #root = 1 to Numvars.
do if ( evalsr(#root,1) < prcntyle(#root,1) ).
compute nfPARPCA = #root - 1.
break.
end if.
end loop.
compute nfactors = nfPARPCA.
end if.

* Parallel Analysis program for Principal Axis/Common Factors.
do if ( roottest = 3).
compute Ncases   = nrow(data). 
compute Numvars  = nrow(r).
compute rdum1 = r.
compute smc = 1 - (1 &/ diag(inv(rdum1)) ).
call setdiag(rdum1,smc).
compute evalsr = eval(rdum1).
* computing random data correlation matrices & eigenvalues.
compute evals = make(Numvars,ndatsets,-9999).
compute nm1 = 1 / (ncases-1).
loop #nds = 1 to ndatsets.
compute x = sqrt(2 * (ln(uniform(ncases,Numvars)) * -1) ) &*
            cos(6.283185 * uniform(ncases,Numvars) ).
compute vcv = nm1 * (sscp(x) - ((t(csum(x))*csum(x))/ncases)).
compute d = inv(mdiag(sqrt(diag(vcv)))).
compute rdum2 = d * vcv * d.
compute smc = 1 - (1 &/ diag(inv(rdum2)) ).
call setdiag(rdum2,smc).
compute evals(:,#nds) = eval(rdum2).
end loop.
* identifying the eigenvalues corresponding to the desired percentile.
compute num = rnd((percent*ndatsets)/100).
compute prcntyle = { t(1:Numvars) }.
loop #root = 1 to Numvars.
compute ranks = rnkorder(evals(#root,:)).
loop #col = 1 to ndatsets.
do if (ranks(1,#col) = num).
compute prcntyle(#root,1) = evals(#root,#col).
break.
end if.
end loop.
end loop.
compute nfPARPAF = 0.
loop #root = 1 to Numvars.
do if ( evalsr(#root,1) < prcntyle(#root,1) ).
compute nfPARPAF = #root - 1.
break.
end if.
end loop.
compute nfactors = nfPARPAF.
end if.

*  Velicer's Minimum Average Partial (MAP) Test.
do if (roottest = 4).
call eigen (r,eigvect,eigval).
compute loadings = eigvect * sqrt(mdiag(eigval)).
compute fm = make(nrow(r),2,-9999).
compute fm(1,2) = (mssq(r)-ncol(r))/(ncol(r)*(ncol(r)-1)).
loop #m = 1 to ncol(r) - 1.
compute a = loadings(:,1:#m).
compute partcov = r - (a * t(a)).
compute d = mdiag( 1 / (sqrt(diag(partcov))) ).
compute pr = d * partcov * d.
compute fm(#m+1,2) = (mssq(pr)-ncol(r))/(ncol(r)*(ncol(r)-1)).
end loop.
* identifying the smallest fm value & its location (= # factors).
compute minfm = fm(1,2).
compute nfVEL = 0.
loop #s = 1 to nrow(fm).
compute fm(#s,1) = #s -1.
do if ( fm(#s,2) < minfm ).
compute minfm = fm(#s,2).
compute nfVEL = #s - 1.
end if.
end loop.
compute nfactors = nfVEL.
end if.

* Standard Error Scree (Zoski & Jurs, 1996, EPM, p 443).
do if ( roottest = 5).
compute evals = eval(r).
compute nevals = nrow(evals).
compute rootnum = t(1:nevals).
compute sesdata = { rootnum , evals }.
compute nfSCREE = 0.
compute arbiter = 1 / nevals.
loop #loop = 1 to nevals-2.
compute y = sesdata(#loop:nevals,2) .
compute n = nrow(y).
compute x = { make(n,1,1) , sesdata( #loop:nevals ,1) }.
compute b = inv( t(x) * x) * t(x) * y.
compute pred = x * b.
compute sse = msum ( (y - pred)&**2 ).
compute sderest = sqrt( sse / (n - 2) ).
do if ( sderest > arbiter ).
compute nfSCREE = nfSCREE + 1.
end if.
end loop.
compute nfactors = nfSCREE.
end if.

*  # of eigenvalues > 1.
do if ( roottest = 6).
compute evalsr = eval(r).
compute nevalgt1 = 0.
loop #a = 1 to nrow(evalsr).
do if ( evalsr(#a,1) > 1.0).
compute nevalgt1 = nevalgt1 + 1.
end if.
end loop.
compute nfactors = nevalgt1.
end if.

compute warnnf1  = 0.
do if ( nfactors = 0).
compute nfactors = 1.
compute warnnf1  = 1.
end if.


* Factor Extraction.
* CFA / PAF  -- from Bernstein p 189   --  smc = from Bernstein p 104.
do if ( extract = 1).
compute rcfa = r.
compute smc = 1 - (1 &/ diag(inv(rcfa)) ).
loop #a = 1 to itercfa.
call setdiag(rcfa,smc).
call eigen(rcfa,eigvect,eigval).
compute eigval = mdiag(eigval).
compute lding=eigvect(:,1:nfactors)*sqrt(eigval(1:nfactors,1:nfactors)).
compute communal = rssq(lding).
do if ( mmax(abs(communal-smc))<.001 ).
break.
else.
compute smc=communal.
end if.
end loop.
end if.

* PCA.
do if ( extract = 2).
call eigen (r,eigvect,eigval).
do if (nfactors = 1).
compute lding = eigvect(:,1:nfactors) * sqrt( (eigval(1:nfactors,:)) ).
end if.
do if (nfactors > 1).
compute lding = eigvect(:,1:nfactors) * sqrt( mdiag(eigval(1:nfactors,:)) ).
end if.
end if.

* Maximum likelihood Factor analysis algorithm -- Marcus, 1993.
do if ( extract = 3).
compute Rho = r.
compute k=nfactors.
compute p = nrow(Rho).
call eigen(Rho,A,L).
compute L = mdiag(L).
compute A1=A(:,1:k)*sqrt(L(1:k,1:k)).
compute Uni=mdiag(diag(Rho-A1*t(A1))).
compute Rh1=sqrt(inv(Uni))*(Rho-Uni)*sqrt(inv(Uni)).
call eigen(Rh1,A,L).
compute L = mdiag(L).
compute A1=sqrt(Uni)*A(:,1:k)*sqrt(L(1:k,1:k)).
compute check=tolerml.
loop #i=1 to iterml.
compute Uni=mdiag(diag(Rho-A1*t(A1))).
compute Rh1=sqrt(inv(Uni))*(Rho-Uni)*sqrt(inv(Uni)).
call eigen(Rh1,A,L).
compute L = mdiag(L).
compute A2=sqrt(Uni)*A(:,1:k)*sqrt(L(1:k,1:k)).
do if ( mmax(abs(A1-A2))<check ).
break.
end if.
compute A1=A2.
end loop.
compute lding = A1.
end if.

* Image Factor Extraction 
  (Gorsuch, 1983, p 113; Velicer 1974, Educ & Psych Meas, 34, p 564).
do if (extract = 4).
compute d =  mdiag( 1 &/ diag(inv(r)) ).
compute gvv = r + d * inv(r) * d - 2 * d.
compute r = gvv.
compute s = sqrt(d).
compute r2 = inv(s) *  gvv  * inv(s).
call eigen(r2,eigvect,eigval).
compute l = eigvect(:,1:nfactors).
compute dd = sqrt(mdiag(eigval(1:nfactors,:))).
compute lding = s * l * dd.
end if.

compute lding = lding(:,1:nfactors).



* Rotation.

do if ( nfactors = 1).
compute rff = 1.
end if.

do if ( nfactors  > 1).

* Promax rotation -- Marcus, 1993.
do if ( rotate = 1).
* varimax rotation -- Marcus, 1993, in Reyment & Joreskog, 
  Applied Factor Analysis in the Natural Sciences, CUP.
compute b=lding.
compute n = nrow(lding).
compute nf = ncol(lding).
compute hjsq=diag(lding*t(lding))   .
compute hj=sqrt(hjsq).
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V0=Vtemp.
loop #it=1 to 10 . 
loop #i =1 to nf-1 .
compute jl=#i+1 .
loop #j=jl to nf.
compute xj=lding(:,#i) &/ hj.   
compute yj=lding(:,#j) &/ hj.   
compute uj=xj &* xj-yj &* yj.
compute vj=2*xj &* yj.
compute A=csum(uj).
compute bigB=csum(vj).
compute C=t(uj) * uj - t(vj) * vj.
compute D=2 * t(uj) * vj.
compute num=D-2*A*bigB/n.
compute den=C-(A**2-bigB**2)/n.
compute tan4p=num/den.
compute tan4p = num / den .
compute artanp = ( artan(abs(tan4p)) ) / 4.
do if (num>0 and den>0).
compute phi = artanp.
else if (num<0 and den<0).
compute phi =  -1 * (.7854 - artanp)  .
else if (num<0 and den>0).
compute phi =  -1 * artanp.
else if (num>0 and den<0).
compute phi =  .7854 - artanp .
end if.
compute angle = phi * 180 / 3.1415926  .
do if (  abs(phi)>.00001  ).
compute bigXj=cos(phi)*xj+sin(phi)*yj.
compute bigYj=-sin(phi)*xj+cos(phi)*yj.
compute bj1=bigXj &* hj.
compute bj2=bigYj &* hj.
compute b(:,#i)=bj1.
compute b(:,#j)=bj2.
compute lding(:,#i)=b(:,#i).
compute lding(:,#j)=b(:,#j).
end if.
end loop.
end loop.
compute lding=b.
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V=Vtemp.
do if ( abs(V-V0)<.0001 ).
break.
else.
compute V0=V.
end if.
end loop.
compute B = lding.
compute Bstar=B&**ppower.
compute Tr=inv(t(B) *B)*t(B)*Bstar.
compute Tr=Tr*sqrt(inv(mdiag(diag(t(Tr)*Tr)))).
compute Tpp=inv(Tr).
compute Tpp=inv(sqrt(mdiag(diag(Tpp*t(Tpp)))))*Tpp.
compute Sr=B*Tr.
compute Phip=Tpp*t(Tpp).
compute Sp=B*t(Tpp).
compute Pp=B*inv(Tpp).
compute Pr=B*inv(t(Tr)).
compute Phir= t(Tr)*Tr.
compute rff = Phip.
do if (loadmat = 1).
compute lding = Sp.
else if (loadmat = 2).
compute lding = Pp.
end if.
end if.


* varimax rotation -- Marcus, 1993, in Reyment & Joreskog, 
  Applied Factor Analysis in the Natural Sciences, CUP.
do if ( rotate = 2).
compute b=lding.
compute n = nrow(lding).
compute nf = ncol(lding).
compute hjsq=diag(lding*t(lding))   .
compute hj=sqrt(hjsq).
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V0=Vtemp.
loop #it=1 to 10 . 
loop #i =1 to nf-1 .
compute jl=#i+1 .
loop #j=jl to nf.
compute xj=lding(:,#i) &/ hj.   
compute yj=lding(:,#j) &/ hj.   
compute uj=xj &* xj-yj &* yj.
compute vj=2*xj &* yj.
compute A=csum(uj).
compute bigB=csum(vj).
compute C=t(uj) * uj - t(vj) * vj.
compute D=2 * t(uj) * vj.
compute num=D-2*A*bigB/n.
compute den=C-(A**2-bigB**2)/n.
compute tan4p=num/den.
compute tan4p = num / den .
compute artanp = ( artan(abs(tan4p)) ) / 4.
do if (num>0 and den>0).
compute phi = artanp.
else if (num<0 and den<0).
compute phi =  -1 * (.7854 - artanp)  .
else if (num<0 and den>0).
compute phi =  -1 * artanp.
else if (num>0 and den<0).
compute phi =  .7854 - artanp .
end if.
compute angle = phi * 180 / 3.1415926  .
do if (  abs(phi)>.00001  ).
compute bigXj=cos(phi)*xj+sin(phi)*yj.
compute bigYj=-sin(phi)*xj+cos(phi)*yj.
compute bj1=bigXj &* hj.
compute bj2=bigYj &* hj.
compute b(:,#i)=bj1.
compute b(:,#j)=bj2.
compute lding(:,#i)=b(:,#i).
compute lding(:,#j)=b(:,#j).
end if.
end loop.
end loop.
compute lding=b.
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V=Vtemp.
do if ( abs(V-V0)<.0001 ).
break.
else.
compute V0=V.
end if.
end loop.

end if.
end if.


compute nfactrs1 = nfactors.


* for regular extension analysis.
* supermatrix: core variables, extension variables, & the factor loadings.
compute sup = make( (ncol(rdata)+nfactors) , (ncol(rdata)+nfactors), 0 ).
compute sup(1:nrow(rdata),1:ncol(rdata)) = rdata.
compute sup( (nrow(sup)-(nfactors-1)):nrow(sup), 1:nrow(lding) ) =
        t(lding(1:core,:)).


* correlations between extension variables & the factors.
compute ppp = make( nrow(sup) , ncol(sup), 0 ).
compute resid = sup.
loop #lupe  = 1 to core.
loop #luper = #lupe to nrow(sup).
compute ppp(#luper,#lupe) = resid(#luper,#lupe) / sqrt(resid(#lupe,#lupe)).
end loop.
compute resid =  resid - ppp(:,#lupe) * t(ppp(:,#lupe)).
end loop.
compute pvo = ppp(1:core,1:core).
compute peo = ppp((core+1):(core+ext),1:core).
compute pfo = ppp( (core+ext+1):(nrow(ppp)) , 1:core).
compute sef1 = peo * t(pfo).



* for higher order factor/extension analysis.
do if ( kind = 2 and nfactrs1 > 1 and rotate = 1).
compute nfactrs1 = nfactors.
compute lding1 = lding.
compute core1  = core.
compute ext1   = ext.

compute r = rff.
compute evals2 = eval(r).


* Number of Factors.

* number of salient loadings as the basis for determining the # of factors
  Gorsuch 1997 J Pers Assess p 545.
do if ( roottest = 1).
*  # of eigenvalues > 1.
compute evals = eval(r).
compute nevalgt1 = 0.
loop #a = 1 to nrow(evals).
do if ( evals(#a,1) > 1.0).
compute nevalgt1 = nevalgt1 + 1.
end if.
end loop.
compute nfactors = nevalgt1.
* CFA / PAF  -- from Bernstein p 189   --  smc = from Bernstein p 104.
compute rcfa = r.
compute smc = 1 - (1 &/ diag(inv(rcfa)) ).
loop #a = 1 to itercfa.
call setdiag(rcfa,smc).
call eigen(rcfa,eigvect,eigval).
compute eigval = mdiag(eigval).
compute lding=eigvect(:,1:nfactors)*sqrt(eigval(1:nfactors,1:nfactors)).
compute communal = rssq(lding).
do if ( mmax(abs(communal-smc))<.001 ).
break.
else.
compute smc=communal.
end if.
end loop.
* varimax rotation -- Marcus, 1993, in Reyment & Joreskog, 
  Applied Factor Analysis in the Natural Sciences, CUP.
do if (nfactors > 1).
compute b=lding.
compute n = nrow(lding).
compute nf = ncol(lding).
compute hjsq=diag(lding*t(lding))   .
compute hj=sqrt(hjsq).
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V0=Vtemp.
loop #it=1 to 10 . 
loop #i =1 to nf-1 .
compute jl=#i+1 .
loop #j=jl to nf.
compute xj=lding(:,#i) &/ hj.   
compute yj=lding(:,#j) &/ hj.   
compute uj=xj &* xj-yj &* yj.
compute vj=2*xj &* yj.
compute A=csum(uj).
compute bigB=csum(vj).
compute C=t(uj) * uj - t(vj) * vj.
compute D=2 * t(uj) * vj.
compute num=D-2*A*bigB/n.
compute den=C-(A**2-bigB**2)/n.
compute tan4p=num/den.
compute tan4p = num / den .
compute artanp = ( artan(abs(tan4p)) ) / 4.
do if (num>0 and den>0).
compute phi = artanp.
else if (num<0 and den<0).
compute phi =  -1 * (.7854 - artanp)  .
else if (num<0 and den>0).
compute phi =  -1 * artanp.
else if (num>0 and den<0).
compute phi =  .7854 - artanp .
end if.
compute angle = phi * 180 / 3.1415926  .
do if (  abs(phi)>.00001  ).
compute bigXj=cos(phi)*xj+sin(phi)*yj.
compute bigYj=-sin(phi)*xj+cos(phi)*yj.
compute bj1=bigXj &* hj.
compute bj2=bigYj &* hj.
compute b(:,#i)=bj1.
compute b(:,#j)=bj2.
compute lding(:,#i)=b(:,#i).
compute lding(:,#j)=b(:,#j).
end if.
end loop.
end loop.
compute lding=b.
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V=Vtemp.
do if ( abs(V-V0)<.0001 ).
break.
else.
compute V0=V.
end if.
end loop.
end if.
compute rowmax = rmax(abs(lding)).
compute nfSAL = 0.
loop #lupec = 1 to ncol(lding).
compute nsalynts = 0.
loop #luper = 1 to nrow(lding).
do if ( abs(lding(#luper,#lupec)) >= salvalue and abs(lding(#luper,#lupec)) =
     rowmax(#luper,1) ).
compute nsalynts = nsalynts + 1.
end if.
end loop.
do if ( nsalynts >= numsals ).
compute nfSAL = nfSAL + 1.
end if.
end loop.
compute nfactors = nfSAL.
end if.

* Parallel Analysis program for Principal Components.
do if ( roottest = 2).
compute Ncases   = nrow(data). 
compute Numvars  = nrow(r).
compute evalsr = eval(r).
* computing random data correlation matrices & eigenvalues.
compute evals = make(Numvars,ndatsets,-9999).
compute nm1 = 1 / (ncases-1).
loop #nds = 1 to ndatsets.
compute x = sqrt(2 * (ln(uniform(ncases,Numvars)) * -1) ) &*
            cos(6.283185 * uniform(ncases,Numvars) ).
compute vcv = nm1 * (sscp(x) - ((t(csum(x))*csum(x))/ncases)).
compute d = inv(mdiag(sqrt(diag(vcv)))).
compute evals(:,#nds) = eval(d * vcv * d).
end loop.
* identifying the eigenvalues corresponding to the desired percentile.
compute num = rnd((percent*ndatsets)/100).
compute prcntyle = { t(1:Numvars) }.
loop #root = 1 to Numvars.
compute ranks = rnkorder(evals(#root,:)).
loop #col = 1 to ndatsets.
do if (ranks(1,#col) = num).
compute prcntyle(#root,1) = evals(#root,#col).
break.
end if.
end loop.
end loop.
compute nfPARPCA = 0.
loop #root = 1 to Numvars.
do if ( evalsr(#root,1) < prcntyle(#root,1) ).
compute nfPARPCA = #root - 1.
break.
end if.
end loop.
compute nfactors = nfPARPCA.
end if.

* Parallel Analysis program for Principal Axis/Common Factors.
do if ( roottest = 3).
compute Ncases   = nrow(data). 
compute Numvars  = nrow(r).
compute rdum1 = r.
compute smc = 1 - (1 &/ diag(inv(rdum1)) ).
call setdiag(rdum1,smc).
compute evalsr = eval(rdum1).
* computing random data correlation matrices & eigenvalues.
compute evals = make(Numvars,ndatsets,-9999).
compute nm1 = 1 / (ncases-1).
loop #nds = 1 to ndatsets.
compute x = sqrt(2 * (ln(uniform(ncases,Numvars)) * -1) ) &*
            cos(6.283185 * uniform(ncases,Numvars) ).
compute vcv = nm1 * (sscp(x) - ((t(csum(x))*csum(x))/ncases)).
compute d = inv(mdiag(sqrt(diag(vcv)))).
compute rdum2 = d * vcv * d.
compute smc = 1 - (1 &/ diag(inv(rdum2)) ).
call setdiag(rdum2,smc).
compute evals(:,#nds) = eval(rdum2).
end loop.
* identifying the eigenvalues corresponding to the desired percentile.
compute num = rnd((percent*ndatsets)/100).
compute prcntyle = { t(1:Numvars) }.
loop #root = 1 to Numvars.
compute ranks = rnkorder(evals(#root,:)).
loop #col = 1 to ndatsets.
do if (ranks(1,#col) = num).
compute prcntyle(#root,1) = evals(#root,#col).
break.
end if.
end loop.
end loop.
compute nfPARPAF = 0.
loop #root = 1 to Numvars.
do if ( evalsr(#root,1) < prcntyle(#root,1) ).
compute nfPARPAF = #root - 1.
break.
end if.
end loop.
compute nfactors = nfPARPAF.
end if.

*  Velicer's Minimum Average Partial (MAP) Test.
do if (roottest = 4).
call eigen (r,eigvect,eigval).
compute loadings = eigvect * sqrt(mdiag(eigval)).
compute fm = make(nrow(r),2,-9999).
compute fm(1,2) = (mssq(r)-ncol(r))/(ncol(r)*(ncol(r)-1)).
loop #m = 1 to ncol(r) - 1.
compute a = loadings(:,1:#m).
compute partcov = r - (a * t(a)).
compute d = mdiag( 1 / (sqrt(diag(partcov))) ).
compute pr = d * partcov * d.
compute fm(#m+1,2) = (mssq(pr)-ncol(r))/(ncol(r)*(ncol(r)-1)).
end loop.
* identifying the smallest fm value & its location (= # factors).
compute minfm = fm(1,2).
compute nfVEL = 0.
loop #s = 1 to nrow(fm).
compute fm(#s,1) = #s -1.
do if ( fm(#s,2) < minfm ).
compute minfm = fm(#s,2).
compute nfVEL = #s - 1.
end if.
end loop.
compute nfactors = nfVEL.
end if.

* Standard Error Scree (Zoski & Jurs, 1996, EPM, p 443).
do if ( roottest = 5).
compute evals = eval(r).
compute nevals = nrow(evals).
compute rootnum = t(1:nevals).
compute sesdata = { rootnum , evals }.
compute nfSCREE = 0.
compute arbiter = 1 / nevals.
loop #loop = 1 to nevals-2.
compute y = sesdata(#loop:nevals,2) .
compute n = nrow(y).
compute x = { make(n,1,1) , sesdata( #loop:nevals ,1) }.
compute b = inv( t(x) * x) * t(x) * y.
compute pred = x * b.
compute sse = msum ( (y - pred)&**2 ).
compute sderest = sqrt( sse / (n - 2) ).
do if ( sderest > arbiter ).
compute nfSCREE = nfSCREE + 1.
end if.
end loop.
compute nfactors = nfSCREE.
end if.

*  # of eigenvalues > 1.
do if ( roottest = 6).
compute evalsr = eval(r).
compute nevalgt1 = 0.
loop #a = 1 to nrow(evalsr).
do if ( evalsr(#a,1) > 1.0).
compute nevalgt1 = nevalgt1 + 1.
end if.
end loop.
compute nfactors = nevalgt1.
end if.

do if (  roottest = 7 ).
compute nfactors = nfactsho.
end if.

compute warnnf2  = 0.
do if ( nfactors = 0).
compute nfactors = 1.
compute warnnf2  = 1.
end if.


* Factor Extraction.

* CFA / PAF  -- from Bernstein p 189   --  smc = from Bernstein p 104.
do if ( extract = 1).
compute rcfa = r.
compute smc = 1 - (1 &/ diag(inv(rcfa)) ).
loop #a = 1 to itercfa.
call setdiag(rcfa,smc).
call eigen(rcfa,eigvect,eigval).
compute eigval = mdiag(eigval).
compute lding=eigvect(:,1:nfactors)*sqrt(eigval(1:nfactors,1:nfactors)).
compute communal = rssq(lding).
do if ( mmax(abs(communal-smc))<.001 ).
break.
else.
compute smc=communal.
end if.
end loop.
end if.

* PCA.
do if ( extract = 2).
call eigen (r,eigvect,eigval).
do if (nfactors = 1).
compute lding = eigvect(:,1:nfactors) * sqrt( (eigval(1:nfactors,:)) ).
end if.
do if (nfactors > 1).
compute lding = eigvect(:,1:nfactors) * sqrt( mdiag(eigval(1:nfactors,:)) ).
end if.
end if.

* Maximum likelihood Factor analysis algorithm -- Marcus, 1993.
do if ( extract = 3).
compute Rho = r.
compute k=nfactors.
compute p = nrow(Rho).
call eigen(Rho,A,L).
compute L = mdiag(L).
compute A1=A(:,1:k)*sqrt(L(1:k,1:k)).
compute Uni=mdiag(diag(Rho-A1*t(A1))).
compute Rh1=sqrt(inv(Uni))*(Rho-Uni)*sqrt(inv(Uni)).
call eigen(Rh1,A,L).
compute L = mdiag(L).
compute A1=sqrt(Uni)*A(:,1:k)*sqrt(L(1:k,1:k)).
compute check=tolerml.
loop #i=1 to iterml.
compute Uni=mdiag(diag(Rho-A1*t(A1))).
compute Rh1=sqrt(inv(Uni))*(Rho-Uni)*sqrt(inv(Uni)).
call eigen(Rh1,A,L).
compute L = mdiag(L).
compute A2=sqrt(Uni)*A(:,1:k)*sqrt(L(1:k,1:k)).
do if ( mmax(abs(A1-A2))<check ).
break.
end if.
compute A1=A2.
end loop.
compute lding = A1.
end if.

* Image Factor Extraction 
  (Gorsuch, 1983, p 113; Velicer 1974, Educ & Psych Meas, 34, p 564).
do if (extract = 4).
compute d =  mdiag( 1 &/ diag(inv(r)) ).
compute gvv = r + d * inv(r) * d - 2 * d.
compute r = gvv.
compute s = sqrt(d).
compute r2 = inv(s) *  gvv  * inv(s).
call eigen(r2,eigvect,eigval).
compute l = eigvect(:,1:nfactors).
do if ( nfactors > 1).
compute dd = sqrt(mdiag(eigval(1:nfactors,:))).
else if ( nfactors = 1).
compute dd = sqrt(eigval(1:nfactors,:)).
end if.
compute lding = s * l * dd.
end if.

compute lding = lding(:,1:nfactors).



* Rotation.
do if ( nfactors  > 1).

* Promax rotation -- Marcus, 1993.
do if ( rotate = 1).
* varimax rotation -- Marcus, 1993, in Reyment & Joreskog, 
  Applied Factor Analysis in the Natural Sciences, CUP.
compute b=lding.
compute n = nrow(lding).
compute nf = ncol(lding).
compute hjsq=diag(lding*t(lding))   .
compute hj=sqrt(hjsq).
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V0=Vtemp.
loop #it=1 to 10 . 
loop #i =1 to nf-1 .
compute jl=#i+1 .
loop #j=jl to nf.
compute xj=lding(:,#i) &/ hj.   
compute yj=lding(:,#j) &/ hj.   
compute uj=xj &* xj-yj &* yj.
compute vj=2*xj &* yj.
compute A=csum(uj).
compute bigB=csum(vj).
compute C=t(uj) * uj - t(vj) * vj.
compute D=2 * t(uj) * vj.
compute num=D-2*A*bigB/n.
compute den=C-(A**2-bigB**2)/n.
compute tan4p=num/den.
compute tan4p = num / den .
compute artanp = ( artan(abs(tan4p)) ) / 4.
do if (num>0 and den>0).
compute phi = artanp.
else if (num<0 and den<0).
compute phi =  -1 * (.7854 - artanp)  .
else if (num<0 and den>0).
compute phi =  -1 * artanp.
else if (num>0 and den<0).
compute phi =  .7854 - artanp .
end if.
compute angle = phi * 180 / 3.1415926  .
do if (  abs(phi)>.00001  ).
compute bigXj=cos(phi)*xj+sin(phi)*yj.
compute bigYj=-sin(phi)*xj+cos(phi)*yj.
compute bj1=bigXj &* hj.
compute bj2=bigYj &* hj.
compute b(:,#i)=bj1.
compute b(:,#j)=bj2.
compute lding(:,#i)=b(:,#i).
compute lding(:,#j)=b(:,#j).
end if.
end loop.
end loop.
compute lding=b.
compute bh=lding &/ (hj*(make(1,nf,1)))  .
compute Vtemp=n*rsum(csum(bh&**4))-rsum(csum(bh&**2)&**2)  .
compute V=Vtemp.
do if ( abs(V-V0)<.0001 ).
break.
else.
compute V0=V.
end if.
end loop.
compute B = lding.
compute Bstar=B&**ppower.
compute Tr=inv(t(B) *B)*t(B)*Bstar.
compute Tr=Tr*sqrt(inv(mdiag(diag(t(Tr)*Tr)))).
compute Tpp=inv(Tr).
compute Tpp=inv(sqrt(mdiag(diag(Tpp*t(Tpp)))))*Tpp.
compute Sr=B*Tr.
compute Phip=Tpp*t(Tpp).
compute Sp=B*t(Tpp).
compute Pp=B*inv(Tpp).
compute Pr=B*inv(t(Tr)).
compute Phir= t(Tr)*Tr.
do if (loadmat = 1).
compute lding = Sp.
else if (loadmat = 2).
compute lding = Pp.
end if.
end if.

end if.



* supermatrix: core variables, extension variables, & factor loadings.
compute sup = make( (nfactrs1+core+ext+nfactors) ,
              (nfactrs1+core+ext+nfactors), 0 ).

* the core variables for the higher order analysis = 
  the factor intercorrelations.
compute sup(1:nrow(rff),1:ncol(rff)) = rff.

* the first set of extension variables = 
  the variables that were in the first factor analysis.
compute sup( (nfactrs1+1):(nfactrs1+core), 1:(ncol(lding1)) ) = lding1.

* the possible additional extension variables = 
  the variables that were not in the first factor analysis.
do if ( ext > 0).
compute sup( (nfactrs1+core+1):(nfactrs1+core+ext) , 1:nfactors ) = 
        rdata( (core+1):(core+ext), 1:nfactors ).
end if.

* the loadings (correlations) from the higher order factor analysis.
compute sup( (nrow(sup)-(nfactors-1)):nrow(sup), 1:nrow(lding) ) = t(lding).

compute ext = core + ext.
compute core = nfactrs1.


* correlations between extension variables & the higher order factors.
compute ppp = make( nrow(sup) , ncol(sup), 0 ).
compute resid = sup.
loop #lupe  = 1 to core.
loop #luper = #lupe to nrow(sup).
compute ppp(#luper,#lupe) = resid(#luper,#lupe) / sqrt(resid(#lupe,#lupe)).
end loop.
compute resid =  resid - ppp(:,#lupe) * t(ppp(:,#lupe)).
end loop.
compute pvo = ppp(1:core,1:core).
compute peo = ppp((core+1):(core+ext),1:core).
compute pfo = ppp( (core+ext+1):(nrow(ppp)) , 1:core).
compute sef = peo * t(pfo).

end if.


do if ( kind = 1 or (kind = 2 and nfactrs1 = 1) or (kind = 2 and rotate > 1) ).
print { t((1:nrow(evals1))), evals1 } 
      / title='Eigenvalues for the Core Variables:' / format "f7.5".
do if ( warnnf1 = 1).
print / title='WARNING: Zero factors were found in the data.'.
print / title='         The number of factors was therefore set at 1.'.
end if.
print nfactors / title='Number of Factors = '.
do if ( nfactrs1 = 1).
print / title='WARNING: There was only one factor, rotation not performed'.
end if.
compute fs={"F","F","F","F","F","F","F","F","F","F","F","F","F","F","F","F"}.
compute labels = { "Item" , fs(1,1:nfactrs1) }.
do if (rotate = 1 and nfactrs1 > 1).
print { t((1:nrow(rff))), rff } 
      / title='Factor Intercorrelations:'/ format "f5.2" /cnames=labels.
end if.
print { t((1:nrow(lding))), lding }
      / title='Core Variable Loadings on the Factors:'
      / format "f5.2" /cnames=labels.
do if ( ext > 0).
print { t((1:nrow(sef1))), sef1 } 
      / title='Extension Variable Correlations with the Factors'
      / format "f5.2" /cnames=labels.
end if.
end if.


* Warning message.
do if ( kind = 2 and nfactrs1 = 1 ).
print /title='WARNING: Higher Order Extension Factor Analysis was requested,'.
print /title='         but there was only one factor in the lower-order data.'.
print /title='         The higher order analyses were not performed.'.
end if.

* Warning message.
do if (kind = 2 and rotate > 1).
print /title='WARNING: Higher Order Factor/Extension Analysis was requested,'.
print /title='         as was a ROTATE option that does not produce'.
print /title='         correlations between factors.'.
print /title='         The higher order analyses were not performed.'.
end if.


do if ( kind = 2 and nfactrs1 > 1 and rotate = 1 ).
print { t((1:nrow(evals1))), evals1 }
      / title='Eigenvalues for the First Set of Core Variables:'/format "f7.5".
print nfactrs1 / title='Number of Factors in the First Set of Core Variables = '.
print { t((1:nrow(evals2))), evals2 }
      / title='Eigenvalues for the Higher Order Factor Analysis:'/format "f7.5".
do if ( warnnf2  = 1).
print / title='WARNING: Zero factors were found in the higher order data.'.
print / title='         The number of factors was therefore set at 1.'.
end if.
print nfactors 
      / title='Number of Factors for the Higher Order Factor Analysis = '.
do if ( nfactors = 1).
print / title='WARNING: There was only one factor in the higher order data,'.
print / title='         rotation not performed'.
end if.
compute fs={"F","F","F","F","F","F","F","F","F","F","F","F","F","F","F","F"}.
compute hos={"HO","HO","HO","HO","HO","HO","HO","HO"}.
compute labels = { "Item" , fs(1,1:nfactrs1), hos(1,1:nfactors) }.
print / title='Factor Intercorrelations from the First Factor Analysis,'.
print / title='and, in the far-right collumn(s), the Item'.
print / title='Loadings on the Higher Order Factor(s):'.
print { t((1:nrow(lding))), rff, lding} /title=' ' /format "f5.2"/cnames=labels.
print / title='Core Variable Loadings on the Lower Order Factors,'.
print / title='and, in the far-right collumn(s), their'.
print / title='Correlations with the Higher Order Factor(s):'.
print { t((1:nrow(lding1))), lding1, sef(1:core1,:) } 
     / title=' '/ format "f5.2" / cnames=labels.
do if ( ext1 > 0).
print / title='Extension Variable Correlations with the Lower Order Factor(s)'.
print / title='and, in the far-right collumn(s), their Correlations'.
print / title='with the Higher Order Factor(s):' .
print { t((1:ext1)), sef1, sef((core1+1):nrow(sef),:) } 
     / title=' ' / format "f5.2" /cnames=labels.
end if.
end if.

end matrix.








